@@ -1,4 +1,4 @@
-# -*- coding: utf-8 -*-
+# josa
 
 # ***** BEGIN LICENSE BLOCK *****
 # Version: MPL 1.1/GPL 2.0/LGPL 2.1
@@ -37,66 +37,133 @@
 # ***** END LICENSE BLOCK *****
 
 import config
+import encoding
 from flags import *
 from jamo import *
 import suffix
 
 import unicodedata
 
-def nfd(u8str):
-    return unicodedata.normalize('NFD', u8str.decode('UTF-8')).encode('UTF-8')
+if config.internal_encoding == '2+RST':
+    L_ALL = 'ㄱㄲㄴㄷㄸㄹㅁㅂㅃㅅㅆㅇㅈㅉㅊㅋㅌㅍㅎ'
+    T_ALL = 'ㄱㄲㄴㄷㄹㅁㅂㅅㅆㅇㅈㅊㅋㅌㅍㅎ'
+    V_ALL = 'ㅏㅐㅑㅒㅓㅔㅕㅖㅗㅛㅜㅠㅡㅣ'
+    L_HIEUH = 'ㅎ'
+    L_IEUNG = 'ㅇ'
+    L_NIEUN = 'ㄴ'
+    T_HIEUH = 'ㅎ'
+    T_IEUNG = 'ㅇ'
+    T_MIEUM = 'ㅁ'
+    T_NIEUN = 'ㄴ'
+    T_PIEUP = 'ㅂ'
+    T_RIEUL = 'ㄹ'
+    T_RIEUL_MIEUM = 'ㄹㅁ'
+    T_SIOS = 'ㅅ'
+    T_SSANGSIOS = 'ㅆ'
+    T_TIKEUT = 'ㄷ'
+    V_A = 'ㅏ'
+    V_AE = 'ㅐ'
+    V_E = 'ㅔ'
+    V_EO = 'ㅓ'
+    V_EU = 'ㅡ'
+    V_I = 'ㅣ'
+    V_O = 'ㅗ'
+    V_OE = 'ㅗㅣ'
+    V_U = 'ㅜ'
+    V_WA = 'ㅗㅏ'
+    V_WAE = 'ㅗㅐ'
+    V_WEO = 'ㅜㅓ'
+    V_YA = 'ㅑ'
+    V_YAE = 'ㅒ'
+    V_YE = 'ㅖ'
+    V_YEO = 'ㅕ'
+
+
+def ENC(unistr):
+    if config.internal_encoding == '2+RST':
+        return encoding.encode(unistr)
+    else:
+        return unicodedata.normalize('NFD', unistr)
+
+
 def NFD(unistr):
     return unicodedata.normalize('NFD', unistr)
+
+
 def NFC(unistr):
     return unicodedata.normalize('NFC', unistr)
 
-ALPHA_ALL = ''.join(map(unichr, range(ord('0'),ord('9')+1) + range(ord('a'),ord('z')+1)))
 
-## 임의로 허용하는 로마자로 된 단어는 음운 구별을 하지 않는다. 할 방법이 없음.
+ALPHA_ALL = '01234567890abcdefghijklmnopqrstuvwxyz'
+
+#
+# 임의로 허용하는 로마자로 된 단어는 음운 구별을 하지 않는다. 할 방법이 없음.
 COND_ALL = '.'
-COND_V_ALL = '[%s]' % (V_ALL + ALPHA_ALL)
-COND_T_ALL = '[%s]' % (T_ALL + ALPHA_ALL)
-COND_V_OR_RIEUL = '[%s]' % (V_ALL + T_RIEUL + ALPHA_ALL)
-COND_T_NOT_RIEUL = '[%s]' % (T_ALL.replace(T_RIEUL, '') + ALPHA_ALL)
+if config.internal_encoding == '2+RST':
+    COND_V_ALL = '[ㅏㅑㅐㅒㅗㅛㅓㅔㅕㅖㅜㅠㅡㅣ%s]' % (ALPHA_ALL)
+    COND_T_ALL = '[ㄱㄲㄴㄷㄹㅁㅂㅅㅆㅇㅈㅊㅋㅌㅍㅎ%s]' % (ALPHA_ALL)
+    COND_V_OR_RIEUL = '[ㅏㅣㅗㅡㅓㅜㅕㅔㅐㅛㅠㅑㅖㅒㄹ%s]' % (ALPHA_ALL)
+    COND_T_NOT_RIEUL = '[ㄱㄲㄴㄷㅁㅂㅅㅆㅇㅈㅊㅋㅌㅍㅎ%s]' % (ALPHA_ALL)
+else:
+    COND_V_ALL = '[%s%s]' % (V_ALL, ALPHA_ALL)
+    COND_T_ALL = '[%s%s]' % (T_ALL, ALPHA_ALL)
+    COND_V_OR_RIEUL = '[%s%s]' % (V_ALL + T_RIEUL, ALPHA_ALL)
+    COND_T_NOT_RIEUL = '[%s%s]' % (T_ALL.replace(T_RIEUL, ''), ALPHA_ALL)
+
+
+TRYCHARS = ''
+
 
 class JosaClass:
     next_flag = josas_flag_start
+
     def __init__(self, rules=[], after=[], notafter=[]):
         self.rules = rules
         self.after = after
         self.notafter = notafter
         self.flag = JosaClass.next_flag
         JosaClass.next_flag += 1
+
     def match(self, word, pos, props):
+        pos_base = pos.split(':')[0]
+        if pos_base == '특수':
+            pos_base = ':'.join(pos.split(':')[0:2])
+
         if self.notafter:
-            if (word, '#' + pos) in self.notafter:
+            if (word, '#' + pos_base) in self.notafter:
                 return False
         if self.after:
             # 부사는 상태부사, 성상부사, 정도부사, 양태부사만 보조사 허용
-            if pos.startswith('부사'):
+            if pos_base == '부사':
                 if pos in ['부사:상태', '부사:성상', '부사:정도', '부사:양태']:
-                    pos = '부사'
+                    pass
                 else:
                     return False
 
-            if ('#' + pos) in self.after:
+            if ('#' + pos_base) in self.after:
                 return True
-            elif (word, '#' + pos) in self.after:
+            elif ('#' + pos) in self.after:
+                return True
+            elif (word, '#' + pos_base) in self.after:
                 return True
             else:
                 return False
-        return True
+        else:
+            return True
+
     def output(self):
+        if len(self.rules) == 0:
+            return ''
         result = []
         line = 'SFX %d Y %d' % (self.flag, len(self.rules))
-        result.append(nfd(line))
+        result.append(line)
         for (sfx, cond, strip) in self.rules:
             if not strip:
                 strip = '0'
             if not cond:
                 cond = '.'
-            line = 'SFX %d %s %s %s' % (self.flag, strip, sfx, cond)
-            result.append(nfd(line))
+            line = 'SFX %d %s %s %s' % (self.flag, ENC(strip), ENC(sfx), cond)
+            result.append(line)
         return '\n'.join(result)
 
 
@@ -105,116 +172,128 @@ class JosaClass:
 
 groups = {}
 
-## '이' 주격/보격 조사
+#
+# '이' 주격/보격 조사
 groups['이'] = [
-    JosaClass(
-        rules = [(u'이', COND_T_ALL, '')],
-        after = ['#명사', '#대명사'],
+    JosaClass(rules=[('이', COND_T_ALL, '')],
+              after=['#명사', '#대명사',
+                     '#특수:숫자', '#특수:알파벳',
+                     '#특수:수:1', '#특수:수:10', '#특수:수:100', '#특수:수:1000',
+                     '#특수:고유수:1', '#특수:고유수:10',
+                     ]
     ),
     # 대명사 '-거'+'이' -> '게'
     JosaClass(
-        rules = [(V_E, u'거', V_EO)],
-        after = [('거', '#대명사'),
-                 ('그거', '#대명사'),
-                 ('요거', '#대명사'),
-                 ('이거', '#대명사'),
-                 ('저거', '#대명사'),
-                ],
+        rules=[(V_E, ENC('거'), V_EO)],
+        after=[('거', '#대명사'),
+               ('그거', '#대명사'),
+               ('요거', '#대명사'),
+               ('이거', '#대명사'),
+               ('저거', '#대명사'),
+               ],
     ),
 ]
 
 # '가' 주격/보격 조사
 groups['가'] = [
     JosaClass(
-        rules = [(u'가', COND_V_ALL, '')],
-        after = ['#명사', '#대명사',
-                 '#특수:숫자', '#특수:알파벳',
-                 '#특수:수:1', '#특수:수:10', '#특수:수:100', '#특수:수:1000',
-                 '#특수:고유수:1', '#특수:고유수:10',
-                 ],
-        notafter = [('나', '#대명사'),
-                    ('너', '#대명사'),
-                    ('저', '#대명사'),
-                   ],
+        rules=[('가', COND_V_ALL, '')],
+        after=['#명사', '#대명사',
+               '#특수:숫자', '#특수:알파벳',
+               '#특수:수:1', '#특수:수:10', '#특수:수:100', '#특수:수:1000',
+               '#특수:고유수:1', '#특수:고유수:10',
+               ],
+        notafter=[('나', '#대명사'),
+                  ('너', '#대명사'),
+                  ('저', '#대명사'),
+                  ],
     ),
     # 대명사 '나'+'가' -> '내가', '너'+'가' -> '네가', '저'+'가' -> '제가'
     JosaClass(
-        rules = [(V_AE + u'가', V_A, V_A),
-                 (V_E + u'가', V_EO, V_EO),
-                 ],
-        after = [('나', '#대명사'),
-                 ('너', '#대명사'),
-                 ('저', '#대명사'),
-                ],
+        rules=[(V_AE + '가', V_A, V_A),
+               (V_E + '가', V_EO, V_EO),
+               ],
+        after=[('나', '#대명사'),
+               ('너', '#대명사'),
+               ('저', '#대명사'),
+               ],
     ),
     # 대명사 '누구'+'가' -> '누가'
     JosaClass(
-        rules = [(u'가', u'누구', u'구')],
-        after = [('누구', '#대명사')],
+        rules=[('가', ENC('누구'), '구')],
+        after=[('누구', '#대명사')],
     ),
 ]
 
 groups['!대명사+의'] = [
     # 대명사 '나'+'의' -> '내', '너'+'의' -> '네', '저'+'의' -> '제'
     JosaClass(
-        rules = [(V_AE, V_A, V_A),
-                 (V_E, V_EO, V_EO),
-                 (V_E, V_EO, V_EO),
-                ],
-        after = [('나', '#대명사'),
-                 ('너', '#대명사'),
-                 ('저', '#대명사'),
-                ],
+        rules=[(V_AE, V_A, V_A),
+               (V_E, V_EO, V_EO),
+               (V_E, V_EO, V_EO),
+               ],
+        after=[('나', '#대명사'),
+               ('너', '#대명사'),
+               ('저', '#대명사'),
+               ],
     ),
 ]
 
 # '-ㄴ', '-ㄹ' 형태로 줄여진 구어체
 groups['!종성줄임'] = [
     JosaClass(
-        rules = [(T_RIEUL, COND_V_ALL, ''),
-                 (T_NIEUN, COND_V_ALL, '')],
-        after = ['#명사', '#대명사'],
+        rules=[(T_RIEUL, COND_V_ALL, ''),
+               (T_NIEUN, COND_V_ALL, '')],
+        after=['#명사', '#대명사'],
     ),
 ]
 
 # '거'+'로' => '걸로'
 groups['!거+로'] = [
     JosaClass(
-        rules = [(T_RIEUL + u'로' + emph, u'거', '')
-                 for emph in ['', u'는', T_RIEUL, u'도',
-                              u'만',
-                              u'서', u'서는', u'선', u'서도',
-                              u'써', u'써는', u'썬', u'써도',
-                              u'부터', u'부터는', u'부턴', u'부터도']],
-        after = [('거', '#대명사'),
-                 ('그거', '#대명사'),
-                 ('요거', '#대명사'),
-                 ('이거', '#대명사'),
-                 ('저거', '#대명사'),
-                ],
+        rules=[(T_RIEUL + '로' + emph, ENC('거'), '')
+               for emph in ['', '는', T_RIEUL, '도',
+                            '만',
+                            '서', '서는', '선', '서도',
+                            '써', '써는', '썬', '써도',
+                            '부터', '부터는', '부턴', '부터도']],
+        after=[('거', '#대명사'),
+               ('그거', '#대명사'),
+               ('요거', '#대명사'),
+               ('이거', '#대명사'),
+               ('저거', '#대명사'),
+               ],
+    ),
+]
+
+# '이서' => '-서'가 숫자 뒤라면 받침이 있을 경우에도 '이서' 형태로 붙을 수 있음
+groups['이서'] = [
+    JosaClass(
+        rules=[('이서', COND_T_ALL, '')],
+        after=['#특수:고유수:1']
     ),
 ]
 
 # 보조사
 groups['!보조사'] = [
     JosaClass(
-        rules = [('도', COND_ALL, ''),
-                 ('만', COND_ALL, ''),
-                 ('은', COND_T_ALL, ''), ('는', COND_V_ALL, ''),
-                 ('은커녕', COND_T_ALL, ''), ('는커녕', COND_V_ALL, ''),
-                ],
-        after = ['#부사',
-                 '#명사', '#대명사', '#수사',
-                 '#특수:숫자', '#특수:알파벳',
-                 '#특수:수:1', '#특수:수:10', '#특수:수:100', '#특수:수:1000',
-                 '#특수:고유수:1', '#특수:고유수:10',
-                 ],
+        rules=[('도', COND_ALL, ''),
+               ('만', COND_ALL, ''),
+               ('은', COND_T_ALL, ''), ('는', COND_V_ALL, ''),
+               ('은커녕', COND_T_ALL, ''), ('는커녕', COND_V_ALL, ''),
+               ],
+        after=['#부사',
+               '#명사', '#대명사', '#수사',
+               '#특수:숫자', '#특수:알파벳',
+               '#특수:수:1', '#특수:수:10', '#특수:수:100', '#특수:수:1000',
+               '#특수:고유수:1', '#특수:고유수:10'
+               ],
     ),
 ]
 
 groups['*'] = [
     JosaClass(
-        rules = [
+        rules=[
          ('을', COND_T_ALL, ''), ('를', COND_V_ALL, ''),
          ('과', COND_T_ALL, ''), ('와', COND_V_ALL, ''),
          ('과는', COND_T_ALL, ''), ('와는', COND_V_ALL, ''),
@@ -268,7 +347,7 @@ groups['*'] = [
          ('께서는', COND_ALL, ''),
          ('께선', COND_ALL, ''),
          ('께서도', COND_ALL, ''),
-         ('끼리', COND_ALL, ''), # TODO: -끼리: 가산명사에만 붙음
+         ('끼리', COND_ALL, ''),         # TODO: -끼리: 가산명사에만 붙음
          ('나', COND_V_ALL, ''),
          ('대로', COND_ALL, ''),
          ('대로는', COND_ALL, ''),
@@ -300,17 +379,22 @@ groups['*'] = [
          ('서', COND_ALL, ''),           # '~에서' 준말
          ('에', COND_ALL, ''),
          ('에게', COND_ALL, ''),
+         ('에게까지', COND_ALL, ''),      # + 보조사 추가
          ('에게는', COND_ALL, ''),
          ('에겐', COND_ALL, ''),
+         ('에게나', COND_ALL, ''),
          ('에게도', COND_ALL, ''),
          ('에게만', COND_ALL, ''),
          ('에게서', COND_ALL, ''),
          ('에게서는', COND_ALL, ''),
          ('에게서도', COND_ALL, ''),
          ('에게서만', COND_ALL, ''),
+         ('에까지', COND_ALL, ''),
+         ('에나', COND_ALL, ''),         # 에+'나' 보조사
          ('에는', COND_ALL, ''),         # 에+'는' 보조사
          ('에라도', COND_ALL, ''),       # 에+'라도' 보조사
          ('에야', COND_ALL, ''),         # 에+'야' 보조사
+         ('에의', COND_ALL, ''),
          ('엔', COND_ALL, ''),           # '에는' 준말
          ('에나', COND_ALL, ''),         # 에+'나' 보조사
          ('에도', COND_ALL, ''),         # 에+'도' 보조사
@@ -325,7 +409,7 @@ groups['*'] = [
          ('이나', COND_T_ALL, ''),
          ('이든', COND_ALL, ''),
          ('이든지', COND_ALL, ''),
-         ('이야', COND_T_ALL, ''), # '-(이, '')야' 강조
+         ('이야', COND_T_ALL, ''),       # '-(이, '')야' 강조
          ('조차', COND_ALL, ''),
          ('조차도', COND_ALL, ''),
          ('처럼', COND_ALL, ''),
@@ -339,11 +423,11 @@ groups['*'] = [
          ('한테도', COND_ALL, ''),
          ('한테서', COND_ALL, ''),
          ],
-        after = ['#명사', '#대명사', '#수사',
-                 '#특수:숫자', '#특수:알파벳',
-                 '#특수:수:1', '#특수:수:10', '#특수:수:100', '#특수:수:1000',
-                 '#특수:고유수:1', '#특수:고유수:10',
-                 ],
+        after=['#명사', '#대명사', '#수사',
+               '#특수:숫자', '#특수:알파벳',
+               '#특수:수:1', '#특수:수:10', '#특수:수:100', '#특수:수:1000',
+               '#특수:고유수:1', '#특수:고유수:10',
+               ],
     ),
 ]
 
@@ -353,19 +437,25 @@ klasses = []
 for _key in groups.keys():
     klasses += groups[_key]
 
+
 def find_flags(word, pos, props):
     result = []
     for klass in klasses:
         if klass.match(word, pos, props):
             result.append(klass.flag)
-    if (pos == '명사' or pos == '대명사' or pos == '특수:복수접미사' or
-        pos == '특수:알파벳' or pos == '특수:숫자' or
-        pos.startswith('특수:수:')):
+    if pos.startswith('명사'):
+        result.append(josa_ida_flag)
+    elif pos in ['대명사', '특수:복수접미사', '특수:알파벳', '특수:숫자']:
+        result.append(josa_ida_flag)
+    elif pos.startswith('특수:수:'):
+        result.append(josa_ida_flag)
+    elif pos.startswith('특수:고유수:'):
         result.append(josa_ida_flag)
     if (pos == '대명사'):
         result.append(josa_ida_t_flag)
     return result
 
+
 def get_ida_rules(flagaliases):
     ida_josas = []
     ida_josas_t = []
@@ -374,33 +464,41 @@ def get_ida_rules(flagaliases):
     ida_conjugations = suffix.make_all_conjugations('이다', '이다', [])
     for c in ida_conjugations:
         if flagaliases and '/' in c:
-            (word,flags_str) = c.split('/')
+            (word, flags_str) = c.split('/')
             cont_flags = [int(s) for s in flags_str.split(',')]
-            if not cont_flags in flagaliases:
+            if cont_flags not in flagaliases:
                 flagaliases.append(cont_flags)
             c = word + '/%d' % (flagaliases.index(cont_flags) + 1)
-        if (NFD(c.decode('utf-8'))[:2] == NFD(u'여') or
-            NFD(c.decode('utf-8'))[:2] == NFD(u'예')):
+        if ENC(c)[:2] in [ENC('여'), ENC('예')]:
             # '이어' -> '여', '이에' => '예' 줄임형은 받침이 있을 경우에만
             ida_josas.append((c, COND_V_ALL))
         else:
             ida_josas.append((c, COND_ALL))
         # '이' 생략
         # TODO: 받침이 앞의 명사에 붙는 경우 허용 여부 (예: "마찬가집니다")
-        if NFC(c.decode('utf-8'))[0] == u'이':
-            ida_josas.append((NFC(c.decode('utf-8'))[1:], COND_V_ALL))
-        elif NFD(c.decode('utf-8'))[:2] == NFD(u'이'):
-            ida_josas_t.append((NFD(c.decode('utf-8'))[2:], COND_V_ALL))
+        if config.internal_encoding == '2+RST':
+            if c.startswith('이'):
+                ida_josas.append((c[1:], COND_V_ALL))
+            elif c.startswith(NFD('이')):
+                ida_josas_t.append((c[2:], COND_V_ALL))
+        else:
+            if NFC(c)[0] == '이':
+                ida_josas.append((NFC(c)[1:], COND_V_ALL))
+            elif NFD(c)[:2] == NFD('이'):
+                ida_josas_t.append((NFD(c)[2:], COND_V_ALL))
 
-    result = ['SFX %d Y %d' % (josa_ida_flag, len(ida_josas))]
-    for (sfx,cond) in ida_josas:
-        result.append(nfd('SFX %d 0 %s %s' % (josa_ida_flag, sfx, cond)))
+    if len(ida_josas) > 0:
+        result = ['SFX %d Y %d' % (josa_ida_flag, len(ida_josas))]
+        for (sfx, cond) in ida_josas:
+            result.append('SFX %d 0 %s %s' % (josa_ida_flag, ENC(sfx), cond))
 
-    result.append('SFX %d Y %d' % (josa_ida_t_flag, len(ida_josas_t)))
-    for (sfx,cond) in ida_josas_t:
-        result.append(nfd('SFX %d 0 %s %s' % (josa_ida_t_flag, sfx, cond)))
+    if len(ida_josas_t) > 0:
+        result.append('SFX %d Y %d' % (josa_ida_t_flag, len(ida_josas_t)))
+        for (sfx, cond) in ida_josas_t:
+            result.append('SFX %d 0 %s %s' % (josa_ida_t_flag, ENC(sfx), cond))
     return result
-    
+
+
 def get_output(flagaliases):
     result = []
     for klass in klasses:
